/*
   CCD defects reader
   by paul69, v0.04, 2003-01-24, <e2500@narod.ru>

   Supported camera:
      E2500 (tested, except black remapping)
      E995 (tested: reading the hot map)
*/
#define STRICT
#include <windows.h>
#include <commctrl.h>
#pragma hdrstop

#include <stdio.h> // sscanf();
#include "rc.h"
#include "usb_connect.h"

#define program STR(PROGRAM_CAPTION)
#define version "v0.04"

//#define TEST 995 // no need real camera

#ifdef TEST

#include <stdlib.h> // rand();

char* fw = "v585-76";
char* sn = "000000000000";
bool DSC::GetVar(int reg, TResponce& data)
{
   if( reg == 0x1A ) {
      data.alloc( strlen(fw)+1 );
      strcpy( (char*)data.ptr->data, fw );
      return true;
   }
   if( reg == 0x19 ) {
      data.alloc( strlen(sn)+1 );
      strcpy( (char*)data.ptr->data, sn );
      return true;
   }
   return false;
}

bool GetDeviceName(void*, TInquiryData& inquiry)
{
   strcpy((char*)inquiry.VendorId, "NIKON   NIKON DSC E2500 1.00");
   return true;
}

void* OpenDSC(char)
{
   return (void*)0;
}
#endif

#define STR(id) (Language::string_table[Language::id][curr_lang])
struct Language
{
enum { EN, RU, LanguageCount };
enum {MESSAGE_LANGUAGE,
   CLEAR_COMPLETE,
   CLEAR_OK,
   CLEAR_FAULT,
   CLEAR_CCD_DEFECTS,
   CAN_NOT_START_CLEAR,
   CLEAR_WARNING,
   STOP_WARNING,
   READ_WARNING,
   REMAP_WARNING,
   ELAPSED_TIME,
   REMAP_FAULT,
   REMAP_OKX,
   REMAP_OK1,
   REMAP_OK0,
   REMAP_COMPLETE,
   TIME_IS_UP,
   CANNOT_OPEN_CAMERA,
   PROGRAM_CAPTION,
   CAN_NOT_START_REMAP,
   REMAP_CCD_DEFECTS,
   REMAP_CCD_BLACK,
   READ_ELAPSED_TIME,
   NOT_IMPLEMENTED,
   READ_CCD_DEFECTS,
   READ_CCD_BLACK,
   StringCount };

   static const char* string_table[StringCount][LanguageCount];
};

int curr_lang;

const char* Language::string_table[Language::StringCount][Language::LanguageCount] = {
{  "English",
   "Russian" },
{  "Deletion is complete",
   " " },
{  "Deletion is complete successfully",
   "  " },
{  "Cannot finish Deletion",
   "  " },
{  "Deletion of the CCD defects tables.\r\n",
   "  .\r\n"},
{  "Cannot start clear procedure.\r\n",
   "   .\r\n"},
{  "Deletion of the defect pixel's maps in your camera. \r\n"
   "This capability is a little testing. You can use it on you own risk. \r\n"
   "If your camera will be damaged, you may be voiding \r\nthe warranty of the camera.\r\n"
   "Continue?",

   "     . \r\n"
   "      .       . \r\n"
   "      ,      . \r\n"
   "?"},

{  "Stop reading? It may take several seconds.",
   " ?     ." },

{  "Reading of the defect pixel's map from your camera. \r\n"
   "It may take about 5 minutes and can be aborted at any time. \r\n"
   "This capability is a little testing. You can use it on you own risk. \r\n"
   "If your camera will be damaged, you may be voiding \r\nthe warranty of the camera.\r\n"
   "Continue?",

   "         . \r\n"
   "     ,       . \r\n"
   "      .       . \r\n"
   "      ,      . \r\n"
   "?"},

{  "The automatic remapping of the CCD defects will be made now. \r\n"
   "You camera will take a shot with exposure\r\nabout 10 seconds and process it. \r\n"
   "This capability is a little testing. You can use it on you own risk. \r\n"
   "If your camera will be damaged, you may be voiding \r\nthe warranty of the camera.\r\n"
   "Continue?",

   "        . \r\n"
   "        10    . \r\n"
   "      .       . \r\n"
   "      ,      . \r\n"
   "?"},
   
{  "Elapsed time: %u sec\r\n",
   " : %u sec\r\n"},
{  "Remap fault or it is not supported\r\n",
   "      .\r\n"},
{  "Remap complete.\r\nIt found %u defect pixels.\r\n",
   " .\r\n %u  .\r\n"},
{  "Remap complete.\r\nIt found one defect pixel.\r\n",
   " .\r\n   .\r\n"},
{  "Remap complete.\r\nIt not found any defect pixels.\r\n",
   " .\r\n   .\r\n"},
{  "Remap complete.\r\n",
   " .\r\n"},
{  "Expected time is over.\r\n",
   "   .\r\n"},
{  "Can not connect to camera.\r\nCheck USB connection and try again.",
   "   .\r\n   USB   ."},
{  "CCD defects reader " version,
   "   " version},
{  "Can not start remap procedure.\r\n",
   "   .\r\n"},
{  "Remapping the hot pixels.\r\n",
   " \"\" .\r\n"},
{  "Remapping the black pixels.\r\n",
   " \"\" .\r\n"},
{  "Elapsed time: %5u sec (%3u%%).\r\nRead %4u defect pixels.",
   " : %5u  (%3u%%).\r\n %4u  ."},
{  "Not inplemented!\r\nTry again next time, please.",
   " !\r\n   ."},
{  "Reading the hot pixels.\r\n",
   "  \"\" .\r\n"},
{  "Reading the black pixels.\r\n",
   "  \"\" .\r\n"},
};


DSC* dsc;
HINSTANCE hAppInst;
HWND hAppWin;
HWND hMsgWin;
HWND hProgressWin;
HWND hLangWin;

void EnableControls(HWND hDlg, BOOL en);

char buffer[0x10000];

enum { IDT_REMAP = 1, IDT_READ, IDT_CLEAR,

   REMAP_TICK = 500,
   MAX_REMAP = 60*1000/REMAP_TICK,

   READ_TICK = 1000,
   MAX_READ = 52,

   CLEAR_TICK = 500,
   MAX_CLEAR = 10*1000/CLEAR_TICK,
};

bool is_busy;
unsigned start;
char* command;
char* status;
char* message;

int remap_count;
int remap_status;
enum { REMAP_ERROR=-2, REMAP_IN_PROGRESS, REMAP_COMPLETE, REMAP_OK };

DWORD __stdcall StartRemap( void* is_black )
{
   remap_status = REMAP_IN_PROGRESS;

#ifdef TEST
   is_black;
   Sleep( 10000 );
   remap_status = REMAP_OK + (rand()&3);
#else

   dsc->SpecCmd(6, is_black ? 0x1E : 0, 15000);

   for(int i=0; i<30; ++i)
   {
      unsigned responce;

      if( dsc->GetReg(0x20, responce) )
      {
         if( (int)responce >= 0 )
         {
            remap_status = REMAP_OK + responce;
            return TRUE;
         }
         break;
      }
      Sleep(1000);
   }

   remap_status = REMAP_COMPLETE;
#endif
   return TRUE;
}

bool Remap(HWND hDlg, int is_black)
{
   if( is_busy )
      return false;

   if( IDYES != MessageBox( hDlg, STR(REMAP_WARNING), program, MB_YESNOCANCEL|MB_ICONQUESTION ) )
      return false;

   EnableControls( hDlg, FALSE );
   remap_count = 0;

   SendMessage(hProgressWin, PBM_SETRANGE, 0, MAKELPARAM(0, MAX_REMAP) );
   SendMessage(hProgressWin, PBM_SETPOS, 0, 0 );

   start = GetTickCount();

   DWORD id;
   HANDLE h = CreateThread(NULL, 0x10000, StartRemap, (void*)is_black, 0, &id);
   if( h == NULL )
   {
      remap_status = REMAP_ERROR;
      strcpy(command, STR(CAN_NOT_START_REMAP));
      if( hMsgWin ) SetWindowText(hMsgWin, buffer);
   }
   else
   {
      CloseHandle( h );

      is_busy = true;
      remap_status = REMAP_IN_PROGRESS;

      strcpy(command, is_black ? STR(REMAP_CCD_BLACK) : STR(REMAP_CCD_DEFECTS));
      if( hMsgWin ) SetWindowText(hMsgWin, buffer);

      status = command + strlen(command);
      SetTimer(hDlg, IDT_REMAP, REMAP_TICK, NULL);
   }
   return true;
}

int clear_count;
int clear_status;
enum { CLEAR_ERROR=-2, CLEAR_IN_PROGRESS, CLEAR_COMPLETE, CLEAR_OK };

DWORD __stdcall StartClear( void* )
{
   clear_status = CLEAR_IN_PROGRESS;

#ifdef TEST
   Sleep( 10000 );
   clear_status = CLEAR_OK;
#else

   dsc->SetReg(0x20, 0x49);
   dsc->SpecCmd(6, 0x16, 3000);

   for(int i=0; i<30; ++i)
   {
      TResponce responce;

      if( dsc->GetVar(0x20, responce) )
      {
         clear_status = CLEAR_OK;
         return TRUE;
      }
      Sleep(1000);
   }

   clear_status = CLEAR_COMPLETE;
#endif
   return TRUE;
}

bool Clear(HWND hDlg)
{
   if( is_busy )
      return false;

   if( IDYES != MessageBox( hDlg, STR(CLEAR_WARNING), program, MB_YESNOCANCEL|MB_ICONQUESTION ) )
      return false;

   EnableControls( hDlg, FALSE );
   clear_count = 0;

   SendMessage(hProgressWin, PBM_SETRANGE, 0, MAKELPARAM(0, MAX_CLEAR) );
   SendMessage(hProgressWin, PBM_SETPOS, 0, 0 );

   start = GetTickCount();

   DWORD id;
   HANDLE h = CreateThread(NULL, 0x10000, StartClear, NULL, 0, &id);
   if( h == NULL )
   {
      clear_status = CLEAR_ERROR;
      strcpy(command, STR(CAN_NOT_START_CLEAR));
      if( hMsgWin ) SetWindowText(hMsgWin, buffer);
   }
   else
   {
      CloseHandle( h );

      is_busy = true;
      clear_status = CLEAR_IN_PROGRESS;

      strcpy(command, STR(CLEAR_CCD_DEFECTS));
      if( hMsgWin ) SetWindowText(hMsgWin, buffer);

      status = command + strlen(command);
      SetTimer(hDlg, IDT_CLEAR, CLEAR_TICK, NULL);
   }
   return true;
}

int pixel_count;
int read_count;
int read_status;
bool break_flag;
bool pause_flag;
bool show_stop;
enum { READ_START = -3, READ_IN_PROGRESS, READ_ERROR, READ_OK };

void SetStatusLine()
{
   wsprintf( status, STR(READ_ELAPSED_TIME), (GetTickCount() - start + 500)/1000, read_count*100/MAX_READ, pixel_count);
}

DWORD __stdcall StartRead( void* is_black )
{
   int max_pixels = is_black ? 30 : 512;
   pixel_count = 0;
   int i, x, y;

   read_status = READ_IN_PROGRESS;

#ifdef TEST
   int last_x, last_y;
   last_x = last_y = 0;
   for(read_count=0; read_count<MAX_READ; ++read_count)
   {
#if TEST == 995
      if( !is_black ) srand(1);
      int n = is_black ? 10 : 16;
#else
      int n = 10;
#endif
      while( pause_flag ) Sleep(500);
      if( break_flag ) break;
      Sleep( 6000 );

      char data[2048];
      char* s = data;
      s += sprintf( s, is_black ? "CCD_BLACK:\r\n" : "CCD_DEFECT:\r\n" );

      for(i=0; i<n; ++i)
         s += sprintf( s, "%d,%d\r\n", ((unsigned)rand()%(1616u+1))-1 , (unsigned)rand()%1204u );

      int len = strlen(message+2)+2;

      const char* end = strchr(data, '\n');
      if( !end )
      {
         wsprintf( message+len, "Unknown responce format:\r\n%s\r\n", data);
         break;
      }

      for(i=0; i<16; ++i)
      {
         const char* cur = end + 1;
         end = strchr(cur, '\n');
         if( !end ) break;

         x = y = -1;
         sscanf(cur, "%i,%i", &x, &y);

         if( x == -1 )
            break;

            if( i == 0 )
            {
               if( read_count == 1)
               {
                  if( last_x == x && last_y == y )
                     break;
               }
               last_x = x;
               last_y = y;
            }

         len += wsprintf( message+len, "%d\t%d\t%d\r\n", pixel_count+1, x, y );

         if( ++pixel_count == max_pixels )
            break;
      }
      if( i<10 ) break;
   }
#else
   TResponce responce;

   int last_x, last_y;
   last_x = last_y = 0;
   
   for(read_count=0; read_count<MAX_READ; ++read_count)
   {
      while( pause_flag ) Sleep(500);
      if( break_flag ) break;

      dsc->SetReg(0x20, read_count);

      dsc->SpecCmd(6, 0x14, 3000);

      if( break_flag ) break;

      dsc->SetReg(0x20, is_black ? 4 : 3);

      if( dsc->SpecCmd(6, 0x16, 3000) && dsc->GetVar(0x1F, responce) )
      {
         int len = strlen(message+2)+2;

         const char* end = strchr((const char*)responce.ptr->data, '\n');
         if( !end )
         {
            wsprintf( message+len, "Unknown responce format:\r\n%s\r\n", responce.ptr->data);
            break;
         }

         for(i=0; i<16; ++i)
         {
            const char* cur = end + 1;
            end = strchr(cur, '\n');
            if( !end ) break;

            x = y = -1;
            sscanf(cur, "%i,%i", &x, &y);

            if( x == -1 )
               break;

            if( i == 0 )
            {
               if( read_count == 1)
               {
                  if( last_x == x && last_y == y )
                     break;
               }
               last_x = x;
               last_y = y;
            }

            len += wsprintf( message+len, "%d\t%d\t%d\r\n", pixel_count+1, x, y );

            if( ++pixel_count == max_pixels )
               break;
         }
         if( i<10 ) break;
      }
   }
#endif

   read_status = READ_OK;
   return TRUE;
}

bool Stop(HWND hDlg)
{
   pause_flag = true;
   if( IDYES != MessageBox( hDlg, STR(STOP_WARNING), program, MB_YESNOCANCEL|MB_ICONQUESTION ) )
   {
      pause_flag = false;
      return false;
   }

   break_flag = true;
   pause_flag = false;

   return true;
}

bool Read(HWND hDlg, int is_black)
{
   if( is_busy )
      return false;

   if( IDYES != MessageBox( hDlg, STR(READ_WARNING), program, MB_YESNOCANCEL|MB_ICONQUESTION ) )
      return false;

   EnableControls( hDlg, FALSE );
   EnableWindow( GetDlgItem(hDlg, IDCANCEL), TRUE );
   SetFocus( GetDlgItem(hDlg, IDCANCEL) );

   break_flag = false;
   pause_flag = false;
   read_count = 0;
   read_status = READ_START;
   SetTimer(hDlg, IDT_READ, READ_TICK, NULL);
   SendMessage(hProgressWin, PBM_SETRANGE, 0, MAKELPARAM(0, MAX_READ) );
   SendMessage(hProgressWin, PBM_SETPOS, 0, 0 );

   start = GetTickCount();

   strcpy(command, is_black ? STR(READ_CCD_BLACK) : STR(READ_CCD_DEFECTS));
   status = command + strlen(command);

   SetStatusLine();
   message = command + strlen(command);

   message[0] = '\r';
   message[1] = '\n';
   message[2] = '\0';

   if( hMsgWin ) SetWindowText(hMsgWin, buffer);

   DWORD id;
   HANDLE h = CreateThread(NULL, 0x10000, StartRead, (void*)is_black, 0, &id);
   if( h == NULL )
      read_status = READ_ERROR;
   else
   {
      is_busy = true;
      CloseHandle( h );

      show_stop = true;
      SetDlgItemText(hDlg, IDCANCEL, "Stop");
   }
   return true;
}

bool SetLanguage( int new_lang )
{
   if( new_lang >= 0 && new_lang < Language::LanguageCount )
   {
      curr_lang = new_lang;
      SetWindowText( hAppWin, STR(PROGRAM_CAPTION) );
      return true;
   }
   return false;
}

void InitDlg(HWND hDlg)
{
   hMsgWin = GetDlgItem(hDlg, IDC_TABLE);
   hProgressWin = GetDlgItem(hDlg, IDC_PROGRESS);
   hLangWin = GetDlgItem(hDlg, IDC_LANGUAGE);

   for(int i=0; i<Language::LanguageCount; ++i)
   {
      curr_lang = i;
      SendMessage(hLangWin, CB_ADDSTRING, 0, (LPARAM)STR(MESSAGE_LANGUAGE) );
   }
   SetLanguage( Language::EN );
   SendMessage(hLangWin, CB_SETCURSEL, 0, curr_lang );

   TInquiryData inquiry;
   TResponce responce;

   int len = 0;
   if( GetDeviceName(dsc->GetHandle(), inquiry) )
   {
      len += wsprintf(&buffer[len], "%8.8s%16.16s%4.4s\r\n",
         inquiry.VendorId, inquiry.ProductId, inquiry.ProductRevisionLevel);
   }

   if( dsc->GetVar(0x1A, responce) )
   {
      len += wsprintf(&buffer[len], "FIRMWARE: %s\r\n", responce.GetStr() );
   }
   if( dsc->GetVar(0x19, responce) )
   {
      len += wsprintf(&buffer[len], "S/N: %s\r\n", responce.GetStr() );
   }

   command = &buffer[len];

   if( hMsgWin ) SetWindowText(hMsgWin, buffer);
}

BOOL Command(HWND hDlg, int NotifyCode, int id, HWND hCtrl)
{
   NotifyCode;
   hCtrl;

   switch(id)
   {
   case IDCANCEL:
      if( show_stop )
         Stop( hDlg );
      else
         PostMessage(hDlg, WM_CLOSE, 0,0);
      return TRUE;

   case IDC_REMAP_DEFECT: Remap( hDlg, FALSE ); return TRUE;
   case IDC_READ_DEFECT: Read( hDlg, FALSE ); return TRUE;
   case IDC_REMAP_BLACK: Remap( hDlg, TRUE ); return TRUE;
   case IDC_READ_BLACK: Read( hDlg, TRUE ); return TRUE;
   case IDC_DELETE_ALL: Clear( hDlg ); return TRUE;

   case IDC_LANGUAGE:
      if( NotifyCode == CBN_SELCHANGE )
      {
         SetLanguage( SendMessage(hLangWin, CB_GETCURSEL, 0, 0 ) );
         return TRUE;
      }
   default:
      return FALSE;
   }
}

void EnableControls(HWND hDlg, BOOL en)
{
   EnableWindow( hLangWin, en );
   EnableWindow( GetDlgItem(hDlg, IDC_DELETE_ALL), en );
   EnableWindow( GetDlgItem(hDlg, IDC_REMAP_DEFECT), en );
   EnableWindow( GetDlgItem(hDlg, IDC_READ_DEFECT), en );
   EnableWindow( GetDlgItem(hDlg, IDC_REMAP_BLACK), en );
   EnableWindow( GetDlgItem(hDlg, IDC_READ_BLACK), en );
   EnableWindow( GetDlgItem(hDlg, IDCANCEL), en );
}

BOOL CALLBACK MainDlg( HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam )
{
   if( uMsg == WM_INITDIALOG )
   {
      if( !hAppWin )
         hAppWin = hDlg;

      InitDlg(hDlg);
      return TRUE;
   }

   if( uMsg == WM_TIMER )
   {
      if( wParam == IDT_READ )
      {
         if( read_status > READ_IN_PROGRESS )
            read_count = MAX_READ;

         SendMessage(hProgressWin, PBM_SETPOS, read_count, 0 );

         if( read_count == MAX_READ )
         {
            is_busy = false;
            KillTimer( hDlg, IDT_READ );

            if( show_stop )
            {
               show_stop = false;
               SetDlgItemText(hDlg, IDCANCEL, "Close");
            }
            EnableControls(hDlg, TRUE);
         }

         SetStatusLine();
         message[0] = '\r';
         message[1] = '\n';

         if( hMsgWin ) SetWindowText(hMsgWin, buffer);

         return TRUE;
      }

      if( wParam == IDT_REMAP )
      {
         ++remap_count;
         if( remap_status > REMAP_IN_PROGRESS )
            remap_count = MAX_REMAP;

         SendMessage(hProgressWin, PBM_SETPOS, remap_count, 0 );

         if( remap_count == MAX_REMAP )
         {
            is_busy = false;
            KillTimer( hDlg, IDT_REMAP );
            EnableControls(hDlg, TRUE);

            if( remap_status == REMAP_IN_PROGRESS )
               wsprintf( status, STR(TIME_IS_UP) );
            else
            if( remap_status == REMAP_COMPLETE )
               wsprintf( status, STR(REMAP_COMPLETE) );
            else
            if( remap_status == REMAP_OK )
               wsprintf( status, STR(REMAP_OK0) );
            else
            if( remap_status == REMAP_OK+1 )
               wsprintf( status, STR(REMAP_OK1) );
            else
            if( remap_status > REMAP_OK )
               wsprintf( status, STR(REMAP_OKX), remap_status-REMAP_OK );
            else
               wsprintf( status, STR(REMAP_FAULT) );
         }
         else
         {
            wsprintf( status, STR(ELAPSED_TIME), (GetTickCount() - start + 500)/1000);
         }

         if( hMsgWin ) SetWindowText(hMsgWin, buffer);
         return TRUE;
      }

      if( wParam == IDT_CLEAR )
      {
         ++clear_count;
         if( clear_status > CLEAR_IN_PROGRESS )
            clear_count = MAX_CLEAR;

         SendMessage(hProgressWin, PBM_SETPOS, clear_count, 0 );

         if( clear_count == MAX_CLEAR )
         {
            is_busy = false;
            KillTimer( hDlg, IDT_CLEAR );
            EnableControls( hDlg, TRUE );

            if( clear_status == CLEAR_IN_PROGRESS )
               wsprintf( status, STR(TIME_IS_UP) );
            else
            if( clear_status == CLEAR_COMPLETE )
               wsprintf( status, STR(CLEAR_COMPLETE) );
            else
            if( clear_status == CLEAR_OK )
               wsprintf( status, STR(CLEAR_OK) );
            else
               wsprintf( status, STR(CLEAR_FAULT) );
         }
         else
         {
            wsprintf( status, STR(ELAPSED_TIME), (GetTickCount() - start + 500)/1000);
         }

         if( hMsgWin ) SetWindowText(hMsgWin, buffer);
         return TRUE;
      }
      return FALSE;
   }
   if( uMsg == WM_COMMAND )
   {
      return Command(hDlg, HIWORD(wParam), LOWORD(wParam), (HWND) lParam);
   }
   if( uMsg == WM_CLOSE )
   {
      if( hDlg == hAppWin )
         hAppWin = NULL;

      hMsgWin = hProgressWin = hLangWin = NULL;
      EndDialog(hDlg, 0);
      return FALSE;
   }
   return FALSE;
}

int WINAPI WinMain( HINSTANCE hInstance, HINSTANCE, LPSTR, int )
{
   hAppInst = hInstance;

   InitCommonControls();

   DSC Camera = OpenDSC();
   if( !Camera.Okay() )
   {
      MessageBox(NULL, STR(CANNOT_OPEN_CAMERA), program, MB_OK|MB_ICONSTOP);
      return 0;
   }
   dsc = &Camera;
   DialogBox( hAppInst, MAKEINTRESOURCE(IDD_MAIN), NULL, MainDlg);
   dsc = NULL;
   return 0;
}
