/*
   USB connection library
   by paul69, v0.02, 2003-01-23, <e2500@narod.ru>

   Supported camera:
      E995 (tested)
      E2500 (tested)
      E4300 (tested)
      E4500 (tested)
*/
#define STRICT
#include <windows.h>
#pragma hdrstop

#include <stdio.h>
#include "usb_connect.h"

FILE* listing;

////////////////////////////////////////////////////////////////////////////////
//
//
//
//
//

struct TCmd
{
   dword command;
   dword size;
};

struct TCmdUSB
{
   dword command;
   dword _zero;

   byte size_hi;
   byte size_lo;
};

struct TScsiPassThrough
{
    word Length;
    byte ScsiStatus;
    byte PathId;
    byte TargetId;
    byte Lun;
    byte CdbLength;
    byte SenseInfoLength;
    byte DataIn;
    byte _align[3];
    dword DataTransferLength;
    dword TimeOutValue;
    void* DataBuffer;
    dword SenseInfoOffset;

    union
    {
       byte Cdb[16];
       
       // sizeof *this == 128 bytes
       byte buffer[128-7*4];

       TCmd Cmd;
       TCmdUSB USB;
    };
};

struct TScsiCommand
{
   byte command;
   byte data_in;
   word timeout;
};

static TScsiCommand command_table[] =
{
   { 0x12, 1, 12 },
   { 0xE0, 0, 2 },
   { 0xE1, 0, 2 },
   { 0xE2, 1, 50 },
   { 0xE3, 1, 2 },
   { 0xE4, 1, 50 },
   { 0xE5, 0, 2 },
};

int timeout;
const int n_commands = sizeof(command_table) / sizeof(command_table[0]);

enum USB_COMMAND
{
   READ_DEVICE_NAME, //12
   USB_START_IO,     //E0
   USB_SEND_CMD,     //E1
   USB_RCV_RESPONCE, //E2
   USB_RCV_ACK,      //E3
   USB_RCV_SIZE,     //E4
   USB_CLOSE,        //E5
};

#ifndef IOCTL_SCSI_PASS_THROUGH_DIRECT
#define IOCTL_SCSI_PASS_THROUGH_DIRECT 0x0004D014
#endif

bool SendCmd_to_DSC(HANDLE hDSC, USB_COMMAND i_command, void* buf, unsigned size)
{
   TScsiCommand command = command_table[i_command];

   TScsiPassThrough data;
   memset(&data, 0, sizeof(data));

   data.Length = 44;
   //data.ScsiStatus = 0;
   //data.PathId = 0;
   data.TargetId = 1;
   //data.Lun = 0;
   data.CdbLength = 16;
   data.SenseInfoLength = 24;
   data.DataIn = command.data_in;
   data.DataTransferLength = size;
   data.TimeOutValue = timeout ? timeout : command.timeout;
   data.DataBuffer = buf;
   data.SenseInfoOffset = 48;

   if(data.DataIn)
      memset(buf, 0, size);

   if( command.command == 0x12 )
   {
      data.CdbLength = 12;
      data.Cmd.command = command.command;
      data.Cmd.size = size;
   }
   else
   {
      data.USB.command = command.command;
      data.USB.size_lo = (byte)(size);
      data.USB.size_hi = (byte)(size >> 8);
   }

   DWORD BytesReturned;
   return 0 != DeviceIoControl( hDSC, IOCTL_SCSI_PASS_THROUGH_DIRECT,
      &data, sizeof(data), &data, sizeof(data), &BytesReturned, NULL );
}

struct TCmdAck
{
   dword length;
   word code; // 1,2,3
   word code2; // 0,0x9FFF
   dword _unknown[14]; // for sizeof == 64
   byte ack;
};


bool GetDeviceName(HANDLE hDSC, TInquiryData& inquiry)
{
   return SendCmd_to_DSC(hDSC, READ_DEVICE_NAME, &inquiry, sizeof(inquiry));
}

bool CloseCamera(HANDLE hDSC)
{
   return SendCmd_to_DSC(hDSC, USB_CLOSE, NULL, 0);
}

struct TSendStartIO
{
   dword length;
   word code; // 1
   word code2; // 0x9FFF
   dword _unknown[2]; // for sizeof == 16
};

static TSendStartIO _Start_IO = { 16, 1, 0x9FFF };

bool DSC::usbStartIO()
{
   if( !this ) return false;
   return SendCmd_to_DSC(handle, USB_START_IO, &_Start_IO, sizeof(_Start_IO) );
}

bool DSC::usbSendCmd(int type, int reg)
{
   if( !this ) return false;

   struct TSendCmd
   {
      dword length;
      word  code; // 2
      word  code2; // 0
      dword _unknown[14]; // for sizeof == 64

      word magic;
      word size;
      byte command;
      byte reg;
      word check_sum;
   } send;
   
   memset(&send, 0, sizeof(send));

   send.length = sizeof(send);
   send.code = 2;

   send.magic = 0x431B;
   send.size = 2;
   send.command = (byte)type;
   send.reg = (byte)reg;
   send.check_sum = (word)((unsigned)send.command + (unsigned)send.reg);

   return SendCmd_to_DSC(handle, USB_SEND_CMD, &send, send.length);
}

bool DSC::usbSendCmdSetStr(int reg, const char* str)
{
   if( !this ) return false;

   struct TSendCmdSetStr
   {
      dword length;
      word  code; // 2
      word  code2; // 0
      dword _unknown[14]; // for sizeof == 64

      word magic;
      word size;
      byte command;
      byte reg;
      byte data[1];
   };

   int len = strlen(str);
   int size = sizeof(TSendCmdSetStr)-1+len+2;

   TSendCmdSetStr& send = *(TSendCmdSetStr*)new byte[size];
   memset(&send, 0, size);

   send.length = size;
   send.code = 2;

   send.magic = 0x431B;
   send.size = (word)(len+2);
   send.command = 3;
   send.reg = (byte)reg;

   unsigned sum = (unsigned)send.command + (unsigned)send.reg;
   for(int i=0; i<len; ++i)
   {
      send.data[i] = (byte)str[i];
      sum += (unsigned)send.data[i];
   }
   send.data[len] = (byte)(sum);
   send.data[len+1] = (byte)(sum>>8);

   bool ret = SendCmd_to_DSC(handle, USB_SEND_CMD, &send, send.length);
   delete [] (byte*)&send;
   return ret;
}

bool DSC::usbSendCmdSetReg(int reg, unsigned value)
{
   if( !this ) return false;

   struct TSendCmdVar
   {
      dword length;
      word code; // 2
      word code2; // 0
      dword _unknown[14]; // for sizeof == 64

      word magic;
      word size;
      byte command;
      byte reg;
      dword value;
      word check_sum;
   } send;

   memset(&send, 0, sizeof(send));

   send.length = sizeof(send);
   send.code = 2;

   send.magic = 0x431B;
   send.size = 6;
   //send.command = (byte)0;
   send.reg = (byte)reg;
   send.value = value;
   send.check_sum = (word)
    ( (unsigned)(value&0xFF)
    + (unsigned)((value>>8)&0xFF)
    + (unsigned)((value>>16)&0xFF)
    + (unsigned)((value>>24)&0xFF)
    + (unsigned)send.reg);

   return SendCmd_to_DSC(handle, USB_SEND_CMD, &send, send.length);
}

bool DSC::usbSpecCmd(int reg, int value)
{
   if( !this ) return false;

   struct TSendCmd
   {
      dword length;
      word code; // 2
      word code2; // 0
      dword _unknown[14]; // for sizeof == 64

      word magic;
      word size;
      byte command;
      byte reg;
      byte value;
      word check_sum;
   } send;
   
   memset(&send, 0, sizeof(send));

   send.length = sizeof(send);
   send.code = 2;

   send.magic = 0x431B;
   send.size = 3;
   send.command = 2;
   send.reg = (byte)reg;
   send.value = (byte)value;
   send.check_sum = (word)(
      (unsigned)send.command +
      (unsigned)send.reg +
      (unsigned)send.value);

   return SendCmd_to_DSC(handle, USB_SEND_CMD, &send, send.length);
}

bool DSC::usbSendAck()
{
   if( !this ) return false;

   TCmdAck send;
   memset(&send, 0, sizeof(send));

   send.length = sizeof(send);
   send.code = 2;
   send.ack = 6;

   return SendCmd_to_DSC(handle, USB_SEND_CMD, &send, send.length);
}

bool DSC::usbAcceptResponce(void* buf, int size)
{
   if( !this ) return false;
   return SendCmd_to_DSC(handle, USB_RCV_RESPONCE, buf, size);
}

bool DSC::usbGetAck()
{
   if( !this ) return false;

   struct TResponceAck
   {
      dword length;
      dword code;
      word _unknown[3]; // for sizeof == 14
   } responce;

   if( !SendCmd_to_DSC(handle, USB_RCV_ACK, &responce, sizeof(responce)) ) return false;

   return (responce.code == 0x9FFF0003);
}

bool DSC::usbGetSize(unsigned& size)
{
   if( !this ) return false;

   struct TResponceSize
   {
      dword length;
      dword code;
      dword _unknown;
      dword responce_size;
   } responce;

   if( SendCmd_to_DSC(handle, USB_RCV_SIZE, &responce, sizeof(responce) ) )
   {
      size = responce.responce_size;
      return true;
   }
   return false;
}



//#define CHECK(x) if( !(x) ) { return false; }
#define CHECK(x) { if( !(x) ) { if(listing)fprintf(listing," %s:%d %s - fault\n", __FILE__, __LINE__, #x); return false; }}
//#define CHECK(x) { time(); if(listing)fprintf(listing,"%d: %s", __LINE__, #x); if( !(x) ) { elaps(); if(listing)fprintf(listing," - fault\n"); return false; } elaps(); if(listing)fprintf(listing,"\n"); }
#define CHECK2(x) { if(listing)fprintf(listing,"%d: %s\n", __LINE__, #x); if( !(x) ) { if(listing)fprintf(listing," - fault\n"); return false; } }

unsigned start_time;
unsigned prev_time;

void time()
{
   unsigned cur_time = GetTickCount();
   if( start_time == 0 )
      start_time = cur_time;

   if(listing)fprintf(listing,"%8d: ", cur_time - start_time);
   prev_time = cur_time;
}
void elaps()
{
   unsigned cur_time = GetTickCount();
   if(listing)fprintf(listing," (%d ms) ", cur_time - prev_time);
}


bool DSC::GetVar(int reg, TResponce& data)
{
   CHECK( usbStartIO() );
   CHECK( usbSendCmd(4, reg) );
   CHECK( usbGetAck() );

   unsigned size;
   CHECK( usbStartIO() );
   CHECK( usbGetSize(size) );
   CHECK( size != 0 );
   CHECK( data.alloc( size ) );
   CHECK( usbAcceptResponce(data.ptr, data.size));
   CHECK( usbGetAck() );

   CHECK( usbStartIO() );
   CHECK( usbSendAck() );
   CHECK( usbGetAck() );

   return true;
}

bool DSC::SetStr(int reg, const char* str)
{
   CHECK( usbStartIO() );
   CHECK( usbSendCmdSetStr(reg, str) );
   CHECK( usbGetAck() );

   unsigned size;
   TCmdAck responce;

   CHECK( usbStartIO() );
   CHECK( usbGetSize(size) );
   CHECK( size == sizeof(responce) );
   CHECK( usbAcceptResponce(&responce, size) );
   CHECK( usbGetAck() );

   CHECK( responce.ack == 6 );

   return true;
}

bool DSC::GetReg(int reg, unsigned& value)
{
   CHECK( usbStartIO() );
   CHECK( usbSendCmd(1, reg) );
   CHECK( usbGetAck() );

   value = -1;
   unsigned size;

   struct TResponceCmdGetReg
   {
      dword length;
      word code; // 1,2,3
      word code2; // 0,0x9FFF
      dword _unknown[14]; // for sizeof == 64

      word magic;
      word size;
      dword value;
      word check_sum;
   } responce;

   CHECK( usbStartIO() );
   CHECK( usbGetSize(size) );
   CHECK( size <= sizeof(responce) );
   CHECK( usbAcceptResponce(&responce, size) );
   CHECK( usbGetAck() );

   if( size == sizeof(TCmdAck) )
      if(listing)fprintf(listing,"responce=%02X\n", ((TCmdAck*)&responce)->ack);

   CHECK( usbStartIO() );
   CHECK( usbSendAck() );
   CHECK( usbGetAck() );

   CHECK( size == sizeof(responce) );

   value = responce.value;

   return true;
}

bool DSC::SpecCmd(int reg, unsigned value, unsigned timeout)
{
   CHECK( usbStartIO() );
   CHECK( usbSpecCmd(reg, value) );
   CHECK( usbGetAck() );

   unsigned size;
   TCmdAck responce;

   CHECK( usbStartIO() );
   CHECK( usbGetSize(size) );
   CHECK( size == sizeof(responce) );
   CHECK( usbAcceptResponce(&responce, size) );
   CHECK( usbGetAck() );

   CHECK( responce.ack == 6 );

   if( timeout )
      Sleep( timeout );

   return true;
}

bool DSC::SetReg(int reg, unsigned value)
{
   CHECK( usbStartIO() );
   CHECK( usbSendCmdSetReg(reg, value) );
   CHECK( usbGetAck() );

   unsigned size;
   TCmdAck responce;

   CHECK( usbStartIO() );
   CHECK( usbGetSize(size) );
   CHECK( size == sizeof(responce) );
   CHECK( usbAcceptResponce(&responce, size) );
   CHECK( usbGetAck() );

   if( responce.ack == 6 ) return true;

   if(listing)fprintf(listing,"responce=%02X\n", responce.ack);
   return false;
}

DSC::~DSC()
{
   if( handle == INVALID_HANDLE_VALUE )
      return;

   ::CloseCamera( handle );
   ::CloseHandle( handle );
   handle = INVALID_HANDLE_VALUE;
}

static char DSC_DRIVE[] = "\\\\.\\C:";
static char DSC_PATH[] = "C:\\";

void* OpenDSC(char camera_drive)
{
   if( camera_drive != ' ' )
   {
      DSC_DRIVE[4] = camera_drive;

      return (void*)CreateFile(DSC_DRIVE, GENERIC_READ|GENERIC_WRITE, FILE_SHARE_READ|FILE_SHARE_WRITE,
         NULL, OPEN_EXISTING, 0, NULL );
   }

   DWORD drives = GetLogicalDrives();

   for(int i=2; i < 26; ++i)
   {
      if( drives & (1<<i) )
      {
         DSC_DRIVE[4] =
         DSC_PATH[0] = (char)('A' + i);

         if( GetDriveType( DSC_PATH ) != DRIVE_REMOVABLE	) continue;

         HANDLE hDSC = CreateFile(DSC_DRIVE, GENERIC_READ|GENERIC_WRITE, FILE_SHARE_READ|FILE_SHARE_WRITE,
            NULL, OPEN_EXISTING, 0, NULL );

         if( hDSC == INVALID_HANDLE_VALUE ) continue;

         TInquiryData inquiry;
         bool ret = GetDeviceName(hDSC, inquiry);
         CloseCamera(hDSC);

         if( ret
            && memcmp(inquiry.VendorId, "NIKON   ", 8) == 0
            && memcmp(inquiry.ProductId, "NIKON DSC E", 11) == 0 )
         {
            return (void*)hDSC;
         }

         CloseHandle( hDSC );
      }
   }
   return (void*)INVALID_HANDLE_VALUE;
}


